/* Copyright (c) 2007 Jython Developers */
package org.python.modules.bz2;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import org.python.modules.bz2.io.BZip2CompressorInputStream;
import org.python.modules.bz2.io.BZip2CompressorOutputStream;

import org.python.core.ArgParser;
import org.python.core.Py;
import org.python.core.PyList;
import org.python.core.PyObject;
import org.python.core.PyString;
import org.python.core.PyType;
import org.python.core.util.StringUtil;
import org.python.expose.ExposedGet;
import org.python.expose.ExposedMethod;
import org.python.expose.ExposedNew;
import org.python.expose.ExposedType;


/**
 * A bz2 file-like object.
 */
@ExposedType(name = "bz2.BZ2File")
public class BZ2File extends PyObject {

    // XXX: convert to an Enum or just boolean flags (like PyFile)
    private static final int MODE_CLOSED = 0;
    private static final int MODE_READ = 1;
    //private static final int MODE_READ_EOF = 2;
    private static final int MODE_WRITE = 3;

    /** The name of the file */
    private String name;

    /** InputStream for read mode, OutputStream for write mode */
    private InputStream inputStream;
    private OutputStream outputStream;

    /** Read or write mode */
    private int mode;
    
    private boolean universal;

    /** The current position in the stream */
    private long pos = 0;

    public BZ2File() {
        super();
    }

    public BZ2File(PyType subType) {
        super(subType);
    }

    public BZ2File(String name, String mode, int buffering) {
        this(name, mode, buffering, 9);
    }

    public BZ2File(String name, String mode, int buffering, int compresslevel) {
        super();
        BZ2File___init__(name, mode, buffering, compresslevel);
    }

    @ExposedNew
    @ExposedMethod
    final void BZ2File___init__(PyObject[] args, String[] kwds) {
        ArgParser ap = new ArgParser("__init__", args, kwds, new String[] {
                "name", "mode", "buffering", "compresslevel" }, 1);
        String name = ap.getString(0);
        String mode = ap.getString(1, "r");
        int buffering = ap.getInt(2, 0);
        int compresslevel = ap.getInt(3, 9);
        BZ2File___init__(name, mode, buffering, compresslevel);
    }

    private void BZ2File___init__(String name, String mode, int buffering, int compresslevel) {
        if (compresslevel < 1 || compresslevel > 9) {
            throw Py.ValueError("compresslevel must be between 1 and 9");
        }

        //System.out.println("AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA");
        parseMode(mode);
        //System.out.println(">AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA");
        File file = new File(name);
        if (this.mode == MODE_READ) {
            if (!file.canRead()) {
                // error
            }
            try {
                FileInputStream fileStream = new FileInputStream(file);
                //readMagic(fileStream);
                inputStream = new BZip2CompressorInputStream(fileStream);
            }
            catch (IOException ioe) {
                ioe.printStackTrace();
                System.out.println("IOE: " + ioe);
            }
        }
        else if (this.mode == MODE_WRITE) {
            if (!file.canWrite()) {
                // error
            }
            try {
                file.createNewFile();
                outputStream = new BZip2CompressorOutputStream(new FileOutputStream(file));
            }
            catch (IOException ioe) {
                ioe.printStackTrace();
                System.out.println("IOE: " + ioe);
            }
        }

    }

    private void parseMode(String mode) {
        String cleanMode = null;

        for (int i = 0; i < mode.length(); i++) {
            char charAt = mode.charAt(i);
            boolean error = false;
            switch (charAt) {
            case 'r':
            case 'w':
                if (cleanMode != null) {
                    error = true;
                }
                cleanMode = "" + charAt;
                break;
            case 'b':
                break;
            case 'U':
                universal = true;
                break;
            default:
                error = true;
            }
            if (error) {
                throw Py.ValueError("invalid mode char " + charAt);
            }
        }

        this.mode = cleanMode.equals("r") ? MODE_READ : MODE_WRITE;
    }

    @Override
    public PyObject __iter__() {
        return BZ2File___iter__();
    }

    @ExposedMethod
    final PyObject BZ2File___iter__() {
        if (mode == MODE_CLOSED) {
            throw Py.ValueError("I/O operation on closed file");
        }
        return this;
    }

    @Override
    public PyObject __iternext__() {
        return BZ2File___iternext__();
    }

    final PyObject BZ2File___iternext__() {
        if (mode == MODE_CLOSED) {
            throw Py.ValueError("I/O operation on closed file");
        }

        PyString ret = new PyString(readline());
        if (ret.__len__() == 0)
            return null;
        return ret;
    }

    /*
    @Override
    public boolean __nonzero__() {
        return BZ2File___nonzero__();
    }

    final boolean BZ2File___nonzero__() {
        return super.__nonzero__();
    }
    */

    public void close() {
        BZ2File_close();
    }

    @ExposedMethod
    final void BZ2File_close() {
        try {
            switch (mode) {
            case MODE_READ:
                inputStream.close();
                break;
            case MODE_WRITE:
                outputStream.close();
                break;
            }
        }
        catch (IOException ioe) {
            throw Py.IOError(ioe);
        }
        mode = MODE_CLOSED;
    }

    public PyObject next() {
        return BZ2File_next();
    }

    @ExposedMethod
    final PyObject BZ2File_next() {
        PyObject ret = BZ2File___iternext__();
        if (ret == null)
            throw Py.StopIteration("");
        return ret;
    }

    public String read() {
        return BZ2File_read(-1);
    }

    public String read(int size) {
        return BZ2File_read(size);
    }

    @ExposedMethod(defaults = {"-1"})
    final String BZ2File_read(int size) {
        if (mode == MODE_CLOSED) {
            throw Py.ValueError("I/O operation on closed file");
        }
        else if (mode != MODE_READ) {
            throw Py.ValueError("file is not ready for reading");
        }

        if (size == 0) {
            return "";
        }
        try {
            if (size < 0) {
                StringBuffer stringBuf = new StringBuffer();
                byte[] buf = new byte[10240];
                int count;
                while ((count = inputStream.read(buf)) != -1) {
                    //stringBuf.append(new String(buf));
                    stringBuf.append(StringUtil.fromBytes(buf, 0, count));
                    pos += count;
                }
                return stringBuf.toString();
            }
            else {
                byte[] buf = new byte[size];
                int count;
                if ((count = inputStream.read(buf)) == -1) {
                    return "";
                }
                //int count = inputStream.read(buf);
                //if (count == -1) 
                pos += count;
                //return new String(buf);
                return StringUtil.fromBytes(buf, 0, count);
            }
        }
        catch (IOException ioe) {
            throw Py.IOError(ioe);
        }
    }

    public String readline() {
        return BZ2File_readline(-1);
    }

    public String readline(int size) {
        return BZ2File_readline(size);
    }

    @ExposedMethod(defaults = {"-1"})
    final String BZ2File_readline(int size) {
        if (mode == MODE_CLOSED) {
            throw Py.ValueError("I/O operation on closed file");
        }
        else if (mode != MODE_READ) {
            throw Py.ValueError("file is not ready for reading");
        }

        StringBuffer buf = new StringBuffer();
        while (size < 0 || buf.length() < size) {
            int c;
            try {
                c = inputStream.read();
            }
            catch (IOException ioe) {
                throw Py.IOError(ioe);
            }
            if (c < 0) {
                break;
            }
            buf.append((char)c);
            pos += 1;
            if ((char)c == '\n') {
                break;
            }
        }
        return buf.toString();
    }

    public PyObject readlines() {
        return BZ2File_readlines(0);
    }

    public PyObject readlines(int size) {
        return BZ2File_readlines(size);
    }

    @ExposedMethod(defaults = {"0"})
    final PyObject BZ2File_readlines(int size) {
        if (mode == MODE_CLOSED) {
            throw Py.ValueError("I/O operation on closed file");
        }
        else if (mode != MODE_READ) {
            throw Py.ValueError("file is not ready for reading");
        }

        PyList data = new PyList();
        int count = 0;
        do {
            String buf = readline();
            int len = buf.length();
            if (len == 0) {
                break;
            }
            count += len;
            data.append(new PyString(buf));
        }
        while (size == 0 || count <= size);

        return data;
    }

    public void seek(long offset) {
        BZ2File_seek(offset, 0);
    }

    public void seek(long offset, int whence) {
        BZ2File_seek(offset, whence);
    }

    @ExposedMethod(defaults = {"0"})
    final void BZ2File_seek(long offset, int whence) {
        //throw Py.IOError(".seek() not supported)");
        /*
        if (mode == MODE_CLOSED) {
            throw Py.ValueError("I/O operation on closed file");
        }
        else if (mode != MODE_READ) {
            throw Py.ValueError("seek works only while reading");
        }
        */
    }

    public long tell() {
        return BZ2File_tell();
    }

    @ExposedMethod
    final long BZ2File_tell() {
        if (mode == MODE_CLOSED) {
            throw Py.ValueError("I/O operation on closed file");
        }
        return pos;
    }

    public void write(String data) {
        BZ2File_write(data);
    }

    @ExposedMethod
    final void BZ2File_write(String data) {
        if (mode == MODE_CLOSED) {
            throw Py.ValueError("I/O operation on closed file");
        }
        else if (mode != MODE_WRITE) {
            throw Py.ValueError("file is not ready for writing");
        }
        // XXX:
    }

    public void writelines(PyObject sequence) {
        BZ2File_writelines(sequence);
    }

    @ExposedMethod
    final void BZ2File_writelines(PyObject sequence) {
        if (mode == MODE_CLOSED) {
            throw Py.ValueError("I/O operation on closed file");
        }
        else if (mode != MODE_WRITE) {
            throw Py.ValueError("file is not ready for writing");
        }
        // XXX:
    }

    public PyObject xreadlines() {
        return BZ2File_xreadlines();
    }

    @ExposedMethod
    final PyObject BZ2File_xreadlines() {
        // XXX: self, context manager, etc. see PyFile
        return BZ2File___iter__();
    }

    @Override
    public String toString() {
        return BZ2File_toString();
    }

    @ExposedMethod(names = {"__str__", "__repr__"})
    final String BZ2File_toString() {
        return String.format("<bz2.BZ2File object at >", Py.idstr(this));
    }

    @ExposedGet
    public String getMode() {
        // XXX:
        return "";
    }

    @ExposedGet
    public String getName() {
        return name;
    }

    @ExposedGet
    public boolean getClosed() {
        return mode == MODE_CLOSED;
    }
}
